from neuron import h


def switch_units(legacy):
    try:
        h.nrnunit_use_legacy(legacy)
    except:
        pass


def test_mod_legacy():
    s = h.Section()
    ut = h.UnitsTest(s(0.5))
    h.ion_style("na_ion", 1, 2, 1, 1, 0, sec=s)
    switch_units(1)
    h.finitialize()
    names = ["mole", "e", "faraday", "planck", "hbar", "gasconst"]
    legacy_hex_values = [
        "0x1.fe18fef60659ap+78",
        "0x1.7a4e7164efbbcp-63",
        "0x1.78e54cccccccdp+16",
        "0x1.b85f8c5445f02p-111",
        "0x1.18779454e3d48p-113",
        "0x1.0a10624dd2f1bp+3",
    ]
    for i, n in enumerate(names):
        val = eval("ut." + n)
        print("%s = %s (%s)" % (n, str(val.hex()), str(val)))
        legacy_value = float.fromhex(legacy_hex_values[i])
        assert val == legacy_value
    assert ut.avogadro == float.fromhex(legacy_hex_values[0])
    ghk_std = h.ghk(-50, 0.001, 10, 2)
    erev_std = h.nernst(s(0.5).nai, s(0.5).nao, 1)
    assert ut.ghk == ghk_std
    assert ut.erev == erev_std
    switch_units(0)
    h.finitialize()
    ghk_std = h.ghk(-50, 0.001, 10, 2)
    erev_std = h.nernst(s(0.5).nai, s(0.5).nao, 1)
    assert ut.mole != float.fromhex(legacy_hex_values[0])
    assert ut.e * ut.avogadro == ut.faraday
    assert abs(ut.faraday - h.FARADAY) < 1e-10
    assert ut.gasconst == h.R
    assert ut.k * ut.avogadro == ut.gasconst
    assert abs(ut.planck - ut.hbar * 2.0 * h.PI) < 1e-49
    assert ut.avogadro == h.Avogadro_constant
    assert ut.ghk == h.ghk(-50, 0.001, 10, 2)
    assert ut.erev == h.nernst(s(0.5).nai, s(0.5).nao, 1)


def test_hoc_legacy():
    switch_units(1)  # legacy
    print("R = %s" % str(h.R))
    print("FARADAY = %s" % str(h.FARADAY))
    celsius = 6.3
    ghk = h.ghk(30, 0.01, 10, 1)  # nernst requires a Section to get voltage
    print("ghk = %s" % str(ghk))

    assert h.R == 8.31441
    assert h.FARADAY == 96485.309
    assert ghk == -483.7914803097116

    switch_units(0)  # Modern
    print("R = %s" % str(h.R))
    print("FARADAY = %s" % str(h.FARADAY))
    ghk = h.ghk(30, 0.01, 10, 1)
    print("ghk = %s" % str(ghk))
    assert h.R == 8.31446261815324
    assert ghk == -483.8380971405879


def test_env_legacy():
    for i in [0, 1]:
        a = None
        try:  # test NRNUNIT_USE_LEGACY not whether we can successfully run a subprocess
            import sys, subprocess

            a = (
                "NRNUNIT_USE_LEGACY=%d %s -c 'from neuron import h; print (h.nrnunit_use_legacy())'"
                % (i, sys.executable)
            )
            a = subprocess.check_output(a, shell=True)
            a = int(float(a.decode().split()[0]))
        except:
            pass
        assert a == i


if __name__ == "__main__":
    test_mod_legacy()
    test_hoc_legacy()
    test_env_legacy()



from .rxdException import RxDException

# TODO: This option is not currently observed
use_reaction_contribution_to_jacobian = True

concentration_nodes_3d = "surface"

# how far inside must a voxel be to be consider inside
# the value is necessary to account for floating point precision
ics_distance_threshold = -1e-12

# resolution (relative to dx) of sampling points use to determine surface
# volume fractions.
ics_partial_volume_resolution = 2

# resolution (relative to dx) of sampling points use to determine surface
# areas of surface voxels.
ics_partial_surface_resolution = 1

# the number of electrophysiology fixed steps per rxd step
# WARNING: setting this to anything other than 1 is probably a very bad
#          idea, numerically speaking, at least for now
fixed_step_factor = 1


class _OverrideLockouts:
    def __init__(self):
        self._extracellular = True

    @property
    def extracellular(self):
        return self._extracellular

    @extracellular.setter
    def extracellular(self, val):
        self._extracellular = val


enable = _OverrideLockouts()



import bisect
import numpy
from .geometry3d.graphicsPrimitives import Cylinder, Cone
from neuron import h

# TODO: remove indirection (e.g. use _h_x3d instead of h.x3d)


def centroids_by_segment(sec):
    """
    given a section, returns a dictionary whose entries are lists of cylinders
    of radius 0 that should be used for distance calculations, keyed by section

    .. warning::

        Does not currently support non-frustum based sections (i.e. no support
        for new 3d styles, like soma outlines)

    .. warning::

        This assumes a 3d style exists. The safest way to call this is to call
        h.define_shape() first
    """
    # TODO: fix the issue described in the warning
    #       (when this was written, these objects were only under development)

    n3d = sec.n3d()
    length = sec.L

    arc3d = [sec.arc3d(i) for i in range(n3d)]
    x3d = numpy.array([sec.x3d(i) for i in range(n3d)])
    y3d = numpy.array([sec.y3d(i) for i in range(n3d)])
    z3d = numpy.array([sec.z3d(i) for i in range(n3d)])
    diam3d = numpy.array([sec.diam3d(i) for i in range(n3d)])

    dx = length / sec.nseg
    objs = {}
    for i in range(sec.nseg):
        x_lo = i * dx
        x_hi = (i + 1) * dx
        pts = [x_lo] + _values_strictly_between(x_lo, x_hi, arc3d) + [x_hi]
        local_x3d = numpy.interp(pts, arc3d, x3d)
        local_y3d = numpy.interp(pts, arc3d, y3d)
        local_z3d = numpy.interp(pts, arc3d, z3d)
        local_diam3d = numpy.interp(pts, arc3d, diam3d)
        local_objs = []
        for j in range(len(pts) - 1):
            x0, y0, z0, r0 = (
                local_x3d[j],
                local_y3d[j],
                local_z3d[j],
                local_diam3d[j] / 2.0,
            )
            x1, y1, z1, r1 = (
                local_x3d[j + 1],
                local_y3d[j + 1],
                local_z3d[j + 1],
                local_diam3d[j + 1] / 2.0,
            )
            if x0 != x1 or y0 != y1 or z0 != z1:
                local_objs.append(Cylinder(x0, y0, z0, x1, y1, z1, 0))
        objs[sec((i + 0.5) / sec.nseg)] = local_objs
    return objs


def objects_by_segment(sec):
    """
    given a section, returns a dictionary whose entries are lists of objects
    that should be used for distance calculations, keyed by section

    .. warning::

        Does not currently support non-frustum based sections (i.e. no support
        for new 3d styles, like soma outlines)

    .. warning::

        This assumes a 3d style exists. The safest way to call this is to call
        h.define_shape() first
    """
    # TODO: fix the issue described in the warning
    #       (when this was written, these objects were only under development)

    n3d = sec.n3d()
    length = sec.L

    arc3d = [sec.arc3d(i) for i in range(n3d)]
    x3d = numpy.array([sec.x3d(i) for i in range(n3d)])
    y3d = numpy.array([sec.y3d(i) for i in range(n3d)])
    z3d = numpy.array([sec.z3d(i) for i in range(n3d)])
    diam3d = numpy.array([sec.diam3d(i) for i in range(n3d)])

    dx = length / sec.nseg
    objs = {}
    for i in range(sec.nseg):
        x_lo = i * dx
        x_hi = (i + 1) * dx
        pts = [x_lo] + _values_strictly_between(x_lo, x_hi, arc3d) + [x_hi]
        local_x3d = numpy.interp(pts, arc3d, x3d)
        local_y3d = numpy.interp(pts, arc3d, y3d)
        local_z3d = numpy.interp(pts, arc3d, z3d)
        local_diam3d = numpy.interp(pts, arc3d, diam3d)
        local_objs = []
        for j in range(len(pts) - 1):
            x0, y0, z0, r0 = (
                local_x3d[j],
                local_y3d[j],
                local_z3d[j],
                local_diam3d[j] / 2.0,
            )
            x1, y1, z1, r1 = (
                local_x3d[j + 1],
                local_y3d[j + 1],
                local_z3d[j + 1],
                local_diam3d[j + 1] / 2.0,
            )
            if r0 == r1:
                local_objs.append(Cylinder(x0, y0, z0, x1, y1, z1, r0))
            else:
                local_objs.append(Cone(x0, y0, z0, r0, x1, y1, z1, r1))
        objs[sec((i + 0.5) / sec.nseg)] = local_objs
    return objs


def _values_between(lo, hi, data):
    i_lo = bisect.bisect_left(data, lo)
    i_hi = bisect.bisect_right(data, hi)
    return data[i_lo:i_hi]


def _values_strictly_between(lo, hi, data):
    temp = _values_between(lo, hi, data)
    if temp and temp[0] == lo:
        temp = temp[1:]
    if temp and temp[-1] == hi:
        temp = temp[:-1]
    return temp



try:
    from xml.etree import cElementTree as etree
except ImportError:
    from xml.etree import ElementTree as etree

from . import xml2nrn

# module names derived from the namespace. Add new tags in proper namespace
from . import neuroml
from . import metadata
from . import morphml
from . import biophysics


class FileWrapper:
    def __init__(self, source):
        self.source = source
        self.lineno = 0

    def read(self, bytes):
        s = self.source.readline()
        self.lineno += 1
        return s


# for each '{namespace}element' call the corresponding module.func
def handle(x2n, fw, event, node):
    tag = node.tag.split("}")
    # hopefully a namespace token corresponding to an imported module name
    ns = tag[0].split("/")[-2]
    tag = ns + "." + tag[1]  # namespace.element should correspond to module.func
    f = None
    try:
        if event == "start":
            f = eval(tag)
        elif event == "end":
            f = eval(tag + "_end")
    except:
        pass
    if f:
        x2n.locator.lineno = fw.lineno
        try:
            f(x2n, node)  # handle the element when it opens
        except:
            print(tag, " failed at ", x2n.locator.getLineNumber())
    elif event == "start":
        print("ignore", node.tag)  # no function to handle the element
        return 0
    return 1


def rdxml(fname, ho=None):
    f = FileWrapper(open(fname))
    x2n = xml2nrn.XML2Nrn()
    ig = None
    for event, elem in etree.iterparse(f, events=("start", "end")):
        if ig != elem:
            if handle(x2n, f, event, elem) == 0:
                ig = elem
    if ho:
        ho.parsed(x2n)


if __name__ == "__main__":
    rdxml("temp.xml")



import sys


class coreneuron(object):
    """
    CoreNEURON configuration values.

    An instance of this class stands in as the `neuron.coreneuron` module. Using
    a class instead of a module allows getter/setter methods to be used, which
    lets us ensure its properties have consistent types and values.

    Attributes
    ----------
    cell_permute
    enable
    file_mode
    gpu
    prcellstate
    verbose
    warp_balance

    Examples
    --------
    >>> from neuron import coreneuron
    >>> coreneuron.enable = True
    >>> coreneuron.enable
    True
    """

    def __init__(self):
        self._enable = False
        self._gpu = False
        self._num_gpus = 0
        self._file_mode = False
        self._cell_permute = None
        self._warp_balance = 0
        self._verbose = 2  # INFO
        self._prcellstate = -1
        self._model_stats = False

    def _valid_cell_permute(self):
        return {1, 2} if self._gpu else {0, 1}

    def _default_cell_permute(self):
        return 1 if self._gpu else 0

    @property
    def enable(self):
        """Use CoreNEURON when calling ParallelContext.psolve(tstop)."""
        return self._enable

    @enable.setter
    def enable(self, value):
        self._enable = bool(int(value))

    @property
    def gpu(self):
        """Activate GPU computation in CoreNEURON."""
        return self._gpu

    @gpu.setter
    def gpu(self, value):
        self._gpu = bool(int(value))
        # If cell_permute has been set to a value incompatible with the new GPU
        # setting, change it and print a warning.
        if (
            self._cell_permute is not None
            and self._cell_permute not in self._valid_cell_permute()
        ):
            print(
                "WARNING: coreneuron.gpu = {} forbids coreneuron.cell_permute = {}, using {} instead".format(
                    self._gpu, self._cell_permute, self._default_cell_permute()
                )
            )
            self._cell_permute = self._default_cell_permute()

    @property
    def num_gpus(self):
        """Get/set the number of GPU to use.
        0 means to use all that are avilable.
        """
        return self._num_gpus

    @num_gpus.setter
    def num_gpus(self, value):
        self._num_gpus = int(value)

    @property
    def file_mode(self):
        """Run via file transfer mode instead of in-memory transfer."""
        return self._file_mode

    @file_mode.setter
    def file_mode(self, value):
        self._file_mode = bool(value)

    @property
    def cell_permute(self):
        """Get/set the cell permutation order. Possible values are:
        - 0: no permutation (requires gpu = False)
        - 1: optimize node adjacency
        - 2: optimise parent node adjacency (requires gpu = True)
        """
        if self._cell_permute is None:
            return self._default_cell_permute()
        return self._cell_permute

    @cell_permute.setter
    def cell_permute(self, value):
        value = int(value)
        assert value in self._valid_cell_permute()
        self._cell_permute = value

    @property
    def warp_balance(self):
        """Number of warps to balance. A value of 0 indicates no balancing
        should be done. This has no effect without cell_permute = 2 (and
        therefore gpu = True)."""
        return self._warp_balance

    @warp_balance.setter
    def warp_balance(self, value):
        value = int(value)
        assert value >= 0
        self._warp_balance = value

    @property
    def verbose(self):
        """Configure the verbosity of CoreNEURON. Possible values are:
        - 0: quiet/no output
        - 1: error
        - 2: info (default)
        - 3: debug
        """
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        value = int(value)
        assert value >= 0 and value <= 3
        self._verbose = value

    @property
    def prcellstate(self):
        """Output prcellstate information for the given gid at t=0 and at tstop.
        A value of -1 indicates no prcellstate output will be generated.
        """
        return self._prcellstate

    @prcellstate.setter
    def prcellstate(self, value):
        value = int(value)
        assert value >= -1
        self._prcellstate = value

    @property
    def model_stats(self):
        """Print number of instances of each mechanism and detailed memory stats."""
        return self._model_stats

    @model_stats.setter
    def model_stats(self, value):
        self._model_stats = bool(value)

    def nrncore_arg(self, tstop):
        """
        Return str that can be used for pc.nrncore_run(str)
        based on user properties and current NEURON settings.
        :param tstop: Simulation time (ms)

        """
        from neuron import h

        arg = ""
        tstop = float(tstop)
        if not self.enable:
            return arg

        # note that this name is also used in C++ file nrncore_write.cpp
        CORENRN_DATA_DIR = "corenrn_data"

        # args derived from user properties
        if self._gpu:
            arg += " --gpu"
            if self._num_gpus:
                arg += " --num-gpus %d" % self._num_gpus
        if self._file_mode:
            arg += " --datpath %s" % CORENRN_DATA_DIR
        arg += " --tstop %g" % tstop
        arg += " --cell-permute %d" % self.cell_permute
        if self._warp_balance > 0:
            arg += " --nwarp %d" % self.warp_balance
        if self._prcellstate >= 0:
            arg += " --prcellgid %d" % self.prcellstate
        arg += " --verbose %d" % self.verbose
        if self._model_stats:
            arg += " --model-stats"

        # args derived from current NEURON settings.
        pc = h.ParallelContext()
        cvode = h.CVode()
        if pc.nhost() > 1:
            arg += " --mpi"
        arg += " --voltage 1000."
        if cvode.queue_mode() & 1:
            arg += " --binqueue"
        spkcompress = int(pc.spike_compress(-1))
        if spkcompress:
            arg += " --spkcompress %g" % spkcompress

        # since multisend is undocumented in NEURON, perhaps it would be best
        # to make the next three arg set from user properties here
        multisend = int(pc.send_time(8))
        if (multisend & 1) == 1:
            arg += " --multisend"
            interval = 2 if multisend & 4 else 1
            arg += " --ms-subintervals %d" % interval
            phases = 2 if multisend & 8 else 1
            arg += " --ms-phases %d" % phases

        return arg


sys.modules[__name__] = coreneuron()



"""
do_test.py -- runs a specific test; does not evaluate it

Robert A McDougal
June 23, 2014
"""


def exec_test(the_file, global_var, local_var):
    global_var["__name__"] = the_file
    exec(compile(open(the_file).read(), the_file, "exec"), global_var, local_var)


def do_test(test_to_run, results_location, num_record=10):
    import os

    # switch to the directory
    initial_path = os.getcwd() + "/"
    the_dir, the_file = os.path.split(test_to_run)
    os.chdir(the_dir)

    from neuron import h

    import itertools

    h.nrnunit_use_legacy(True)
    data = {"record_count": 0, "data": []}
    do_test.data = data
    record_count = 0

    def collect_data():
        """grabs the membrane potential data, h.t, and the rxd state values"""
        from neuron import rxd

        data["record_count"] += 1
        if data["record_count"] > num_record:
            save_and_cleanup()

        all_potentials = [seg.v for seg in itertools.chain.from_iterable(h.allsec())]
        rxd_1d = list(rxd.node._states)
        rxd_3d = []
        rxd_ecs = []
        for sp in rxd.species._all_species:
            s = sp()
            if s and hasattr(s, "_intracellular_instances"):
                for key, ics in sorted(
                    s._intracellular_instances.items(), key=lambda val: val[0]._id
                ):
                    rxd_3d += list(ics.states)
            if s and hasattr(s, "_extracellular_instances"):
                for ecs in sorted(
                    s._extracellular_instances.values(), key=lambda spe: spe._grid_id
                ):
                    rxd_ecs += list(ecs.states.flatten())
        all_rxd = rxd_1d + rxd_3d + rxd_ecs
        local_data = [h.t] + all_potentials + all_rxd

        # remove data before t=0
        if h.t == 0:
            data["data"] = []
            data["record_count"] = 1
        # remove previous record if h.t is the same
        if data["record_count"] > 1 and h.t == data["data"][-len(local_data)]:
            data["record_count"] -= 1
            del data["data"][-len(local_data) :]
        # add new data record
        data["data"].extend(local_data)
        # print correct record length
        if data["record_count"] == 2:
            outstr = "<BAS_RL %i BAS_RL> %s %s" % (
                len(local_data),
                repr(h.t),
                data["record_count"],
            )
            print(outstr)

    def save_and_cleanup():
        import array

        os.chdir(initial_path)
        # save the data
        with open(results_location, "wb") as f:
            array.array("d", data["data"]).tofile(f)

        import sys

        sys.exit(0)

    h.CVode().extra_scatter_gather(0, collect_data)

    # now run it
    exec_test(the_file, globals(), globals())

    # should only get here if very few steps
    save_and_cleanup()


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 3:
        print(
            "needs two arguments: relative path to python test to run, relative path for results"
        )
        sys.exit()
    do_test(sys.argv[1], sys.argv[2])



from neuron import h, hoc
import numpy
import neuron
from .rxdException import RxDException

regions = {}

species = {}
all_reactions = {}

rxd_builder_tab = 1
has_instantiated = False

rxd_builder_left = 20
rxd_builder_width = 550
rxd_builder_top = 100
rxd_builder_height = 400


h.load_file("stdlib.hoc")

default_region = {"morphology": "No Sections", "geometry": "Inside"}


def _instantiate_regions(regions):
    seclist_names, seclist_map = get_sectionlists()
    for name, data in zip(list(regions.keys()), list(regions.values())):
        region_spec = ""
        seclist = data.get("morphology", "No Sections")
        if seclist == "No Sections":
            seclist = "[]"
        elif seclist == "All Sections":
            seclist = "h.allsec()"
        elif seclist in seclist_map:
            seclist = "h.SectionList[%d]" % seclist_map[seclist]
        else:
            raise RxDException("Invalid region: %r" % data)
        geo = data.get("geometry", None)
        # TODO: do this without if statements in a way that lets geometries register themselves
        if geo is None:
            raise RxDException("Missing geometry for region %r" % data)
        if geo == "Inside":
            geo = "inside"
        elif geo == "Fractional Volume":
            geo = (
                "FractionalVolume(volume_fraction=%g, surface_fraction=%g, neighbor_areas_fraction=%g)"
                % (data.get("vf", 1), data.get("sf", 0), data.get("nf", 1))
            )
        elif geo == "Constant 2D Area/Length":
            geo = "FixedPerimeter(%g, on_cell_surface=%r)" % (
                data.get("perim", 1),
                data.get("onsurf", False),
            )
        elif geo == "Constant 3D Vol/Length":
            geo = "FixedCrossSection(%g, surface_area=%g)" % (
                data.get("carea", 1),
                data.get("csurf", 0),
            )
        elif geo == "Membrane":
            geo = "membrane"
        elif geo == "Shell":
            geo = "Shell(lo=%g, hi=%g)" % (data.get("lo", 0), data.get("hi", 1))
        else:
            raise RxDException("Unrecognized geometry: %s" % geo)
        nrn_region = data.get("nrn_region", None)
        command = "%s = rxd.Region(%s, geometry=rxd.%s, nrn_region=%r)" % (
            name,
            seclist,
            geo,
            nrn_region,
        )
        exec(command, globals())


def _instantiate_species(species):
    for name, data in zip(list(species.keys()), list(species.values())):
        regions = data.get("regions", [])
        regions = "[%s]" % ",".join(regions)
        command = '%s = rxd.Species(%s, d=%g, name="%s", charge=%g)' % (
            name,
            regions,
            data.get("d", 0),
            name,
            data.get("charge", 0),
        )
        exec(command, globals())


def _construct_side(items):
    result = []
    for item in items:
        if item[1] == 1:
            result.append(item[0])
        elif item[1] != 0:
            result.append("%g * %s" % (item[1], item[0]))
    return " + ".join(result)


def _construct_schema(lhs, rhs):
    return "%s <> %s" % (_construct_side(lhs), _construct_side(rhs))


def _instantiate_reactions(reactions):
    for name, data in zip(list(reactions.keys()), list(reactions.values())):
        reaction_spec = ""
        if "type" not in data:
            raise RxDException("No reaction type for: %r" % data)
        if "active" not in data or not data["active"]:
            continue
        kind = data["type"].lower()
        if kind == "rate":
            reaction_spec = "Rate(%s, %s)" % (data["species"], data["kf"])
        elif kind == "multicompartmentreaction":
            reaction_spec = (
                "MultiCompartmentReaction(%s, %s, %s, membrane=%s, custom_dynamics=%r, membrane_flux=%r, scale_by_area=%r)"
                % (
                    _construct_schema(data["lhs"], data["rhs"]),
                    data.get("kf", 0),
                    data.get("kb", 0),
                    data["membrane"],
                    not data.get("massaction", True),
                    data.get("membrane_current", False),
                    data.get("scale_with_area", True),
                )
            )
        elif kind == "reaction":
            reaction_spec = "Reaction(%s, %s, %s, custom_dynamics=%r)" % (
                _construct_schema(data["lhs"], data["rhs"]),
                data.get("kf", 0),
                data.get("kb", 0),
                not data.get("massaction", True),
            )
        command = "%s = rxd.%s" % (name, reaction_spec)
        exec(command, globals())


def _instantiate():
    global has_instantiated
    if has_instantiated:
        h.continue_dialog("Cannot reinstantiate yet.")
    else:
        has_instantiated = True
        exec("from neuron import rxd", globals())
        _instantiate_regions(regions)
        _instantiate_species(species)
        _instantiate_reactions(all_reactions)


def get_sectionlists():
    found_sectionlists = {}
    results = []
    result_mapping = {}

    # start with the HOC names
    for item in dir(h):
        if item == "nseg":
            continue
        try:
            name = getattr(h, item).hname()
        except:
            continue
        if name[:12] != "SectionList[":
            continue
        found_sectionlists[name] = 0
        result_mapping[item] = int(name[12:].split("]")[0])
        results.append(item)

    # add Python names, if any
    # TODO: is there a better place to look besides globals?
    g = globals()
    for item, value in zip(list(g.keys()), list(g.values())):
        if type(value) == hoc.HocObject:
            try:
                name = value.hname()
            except:
                pass
            else:
                if name[:12] == "SectionList[":
                    found_sectionlists[name] = 0
                    result_mapping[item] = int(name[12:].split("]")[0])
                    results.append(item)

    # sort these
    results = sorted(results, key=lambda s: s.lower())

    # end with the anonymous lists (this probably includes things from cellbuilder)

    # handle any sectionlists without good hoc names
    ell = h.List("SectionList")
    for i in range(int(ell.count())):
        name = ell.object(i).hname()
        if name not in found_sectionlists:
            result_mapping[name] = int(name[12:].split("]")[0])
            results.append(name)

    return results, result_mapping


class _PartialSelector:
    def select(self, name):
        try:
            i = self.names.index(name)
            self.ell.select(i)
            self.currently_selected = name
        except:
            self.ell.select(-1)

    def selected(self):
        i = int(self.ell.selected())
        if i == -1:
            return None
        return self.names[i]

    def select_action(self, action):
        self.ell.select_action(action)


# this is a way of faking the Singleton pattern
def ReactionEditor():
    _the_reaction_editor = RxDBuilder()._the_reaction_editor
    if not _the_reaction_editor.is_mapped:
        _the_reaction_editor.map()
    return _the_reaction_editor


_the_morphology_pane = None


def MorphologyPane():
    global _the_morphology_pane
    if _the_morphology_pane is None:
        _the_morphology_pane = _MorphologyPane()
    _the_morphology_pane.map()
    return _the_morphology_pane


def RxDBuilder(left=None, top=None, width=None, height=None, visible=True):
    global _the_rxd_builder, rxd_builder_left, rxd_builder_width, rxd_builder_height, rxd_builder_top
    if left is not None:
        rxd_builder_left = left
    if top is not None:
        rxd_builder_top = top
    if width is not None:
        rxd_builder_width = width
    if height is not None:
        rxd_builder_height = height
    if _the_rxd_builder is None:
        _the_rxd_builder = _RxDBuilder()
    if not _the_rxd_builder.is_mapped and visible:
        _the_rxd_builder.map()
    return _the_rxd_builder


class FractionalVolumeOptions:
    def __init__(self):
        self.set_options({})
        h.xpanel("")
        h.xvalue("Volume Fraction", (self, "vf"))
        h.xvalue("Surface Fraction", (self, "sf"))
        h.xvalue("Neighbor Fraction", (self, "nf"))
        h.xpanel()

    def get_options(self):
        return {"vf": self.vf, "nf": self.nf, "sf": self.sf}

    def set_options(self, opt):
        self.vf = opt.get("vf", 1)
        self.nf = opt.get("nf", 1)
        self.sf = opt.get("sf", 1)

    info = (
        "Fractional Volume is used to represent regions that are intermixed\n"
        "in a nontrivial way.\n"
        "\n"
        "It is likely that in most cases the volume fraction and the\n"
        "neighbor fraction will both represent the cross-sectional area\n"
        "fraction, and so therefore should be equal. The surface fraction\n"
        "is the fraction of the surface area that belongs to this region.\n"
        "For example, if this is used to represent the ER, then the\n"
        "surface fraction should be zero."
    )


class GeoSelector:
    text = {
        1: "Inside",
        2: "Membrane",
        3: "Fractional Volume",
        4: "Shell",
        5: "Constant 2D Area/Length",
        6: "Constant 3D Vol/Length",
    }

    def select_by_name(self, name):
        click_map = {
            "Inside": self.click_inside,
            "Membrane": self.click_membrane,
            "Fractional Volume": self.click_fractional,
            "Shell": self.click_shell,
            "Constant 2D Area/Length": self.click_constarea,
            "Constant 3D Vol/Length": self.click_constvol,
        }
        click_map.get(name, self.clear_checkboxes)()

    def clear_checkboxes(self):
        self.inside_checkbox = 0
        self.membrane_checkbox = 0.0
        self.fractional_checkbox = 0.0
        self.shell_checkbox = 0.0
        self.constantarea_checkbox = 0.0
        self.constantvol_checkbox = 0.0

    def __init__(self):
        self.clear_checkboxes()
        self.callback = _do_nothing
        self.option = 1
        self.inside_checkbox = 1
        h.xpanel("")
        h.xcheckbox("Inside", (self, "inside_checkbox"), self.click_inside)
        h.xcheckbox("Membrane", (self, "membrane_checkbox"), self.click_membrane)
        h.xcheckbox(
            "Fractional Volume", (self, "fractional_checkbox"), self.click_fractional
        )
        h.xcheckbox("Shell", (self, "shell_checkbox"), self.click_shell)
        h.xcheckbox(
            "Constant 2D Area/Length",
            (self, "constantarea_checkbox"),
            self.click_constarea,
        )
        h.xcheckbox(
            "Constant 3D Vol/Length",
            (self, "constantvol_checkbox"),
            self.click_constvol,
        )
        h.xpanel()

    def click_inside(self):
        self.clear_checkboxes()
        self.inside_checkbox = 1
        self.option = 1
        self.callback()

    def click_membrane(self):
        self.clear_checkboxes()
        self.membrane_checkbox = 1
        self.option = 2
        self.callback()

    def click_fractional(self):
        self.clear_checkboxes()
        self.fractional_checkbox = 1
        self.option = 3
        self.callback()

    def click_shell(self):
        self.clear_checkboxes()
        self.shell_checkbox = 1
        self.option = 4
        self.callback()

    def click_constarea(self):
        self.clear_checkboxes()
        self.constantarea_checkbox = 1
        self.option = 5
        self.callback()

    def click_constvol(self):
        self.clear_checkboxes()
        self.constantvol_checkbox = 1
        self.option = 6
        self.callback()


class InsideOptions:
    def __init__(self):
        h.xpanel("")
        h.xlabel("No options.")
        h.xpanel()

    info = "The interior of the selected portions of the cell."

    def get_options(self):
        return {}

    def set_options(self, opt):
        pass

    is_boundary = False

    def is_valid(self):
        return True


class MembraneOptions:
    def __init__(self):
        h.xpanel("")
        h.xlabel("No options.")
        h.xpanel()

    info = "The membrane of the selected portions of the cell."

    def get_options(self):
        return {}

    def set_options(self, opt):
        pass

    is_boundary = True

    def is_valid(self):
        return True


class ShellOptions:
    def __init__(self):
        self.set_options({})
        h.xpanel("")
        h.xvalue("lo", (self, "lo"))
        h.xvalue("hi", (self, "hi"))
        h.xpanel()

    info = (
        "Shells. lo and hi denote fractional distances along the radius,\n"
        "with the center of the dendrite as 0 and the surface as 1."
    )

    def get_options(self):
        return {"lo": self.lo, "hi": self.hi}

    def set_options(self, opt):
        self.lo = opt.get("lo", 0)
        self.hi = opt.get("hi", 1)

    is_boundary = False

    def is_valid(self):
        return True


class ConstAreaOptions:
    def __init__(self):
        h.xpanel("")
        h.xlabel("Const area options.")
        h.xpanel()

    info = "A fixed membrane size."

    def __init__(self):
        self.set_options({})
        h.xpanel("")
        h.xvalue("Perimeter", (self, "perim"))
        h.xcheckbox("On surface?", (self, "onsurf"))
        h.xpanel()

    def get_options(self):
        return {"perim": self.perim, "onsurf": self.onsurf}

    def set_options(self, opt):
        self.perim = opt.get("perim", 1)
        self.onsurf = opt.get("onsurf", 0)

    is_boundary = True

    def is_valid(self):
        return True


class ConstVolOptions:
    # rxd.FixedCrossSection
    def __init__(self):
        self.set_options({})
        h.xpanel("")
        h.xvalue("Cross-Sectional Area", (self, "carea"))
        h.xvalue("Surface Area/length", (self, "csurf"))
        h.xpanel()

    def get_options(self):
        return {"csurf": self.csurf, "carea": self.carea}

    def set_options(self, opt):
        self.csurf = opt.get("csurf", 0)
        self.carea = opt.get("carea", 1)

    is_boundary = False
    info = (
        "Use constant volume only when you want to neglect diameter effects,\n"
        "such as for point models. In particular, avoid this geometry when the\n"
        "diameter varies."
    )

    def is_valid(self):
        return True


class RegionPane:
    def __init__(self):
        self.hbox1 = h.HBox(3)
        self.hbox1.intercept(1)
        self.vbox1 = h.VBox(3)
        self.vbox1.intercept(1)
        self.region_list = RegionList(None)
        h.xpanel("")
        h.xbutton("Delete", self.delete)
        h.xpanel()
        self.vbox1.intercept(0)
        self.vbox1.map()
        self.vbox2 = h.VBox(3)
        self.vbox2.intercept(1)
        h.xpanel("")
        h.xlabel("Name:")
        h.xpanel()
        self.vbox2.adjuster(15)
        self.name_editor = h.TextEditor("", 1, 30)
        self.name_editor.map()
        self.nrnregion_selector = NrnRegionSelector()
        h.xpanel("")
        h.xlabel("Select Geometry:")
        h.xpanel()
        self.hbox2 = h.HBox(3)
        self.hbox2.intercept(1)
        self.geoselector = GeoSelector()
        self.geoselector.callback = self.geo_switched
        self.deck = h.Deck(3)
        self.deck.intercept(1)
        self.option_constructors = [
            InsideOptions,
            MembraneOptions,
            FractionalVolumeOptions,
            ShellOptions,
            ConstAreaOptions,
            ConstVolOptions,
        ]
        self.option_panels = []
        for panel in self.option_constructors:
            v = h.VBox(3)
            v.intercept(1)
            self.option_panels.append(panel())
            # texteditor persists even if not saved
            spacer = h.TextEditor("", 1, 30)
            spacer.readonly(1)
            spacer.map()
            h.xpanel("")
            h.xbutton("Information", self.info)
            h.xpanel()
            v.intercept(0)
            v.map()
            # vboxes persist even when not saving a reference
        self.deck.intercept(0)
        self.deck.flip_to(0)
        self.deck.map()
        self.hbox2.intercept(0)
        self.hbox2.map()
        h.xpanel("", 1)
        h.xbutton("Revert", self.update)
        h.xbutton("Save", self.save)
        h.xpanel()
        self.vbox2.intercept(0)
        self.vbox2.map()
        self.hbox1.intercept(0)
        self.hbox1.map()
        self.update()
        self.region_list.select_action(self.update)
        self.region_list.select("(NEW)")
        self.update()

    def info(self):
        h.xpanel("Information on %s" % self.geoselector.text[self.geoselector.option])
        option_panel = self.option_panels[self.geoselector.option - 1]
        for line in option_panel.info.split("\n"):
            h.xlabel(line)
        h.xpanel()

    def geo_switched(self):
        # called whenever the geometry has been switched
        geo_type_id = self.geoselector.option - 1
        self.deck.flip_to(geo_type_id)
        region_name = self.region_list.selected()
        region_info = regions.get(region_name, default_region)
        self.option_panels[geo_type_id].set_options(region_info)

    def update(self):
        # update the names
        self.region_list.select_action("")
        selected = self.region_list.selected()
        region_names = sorted(list(regions.keys()), key=lambda s: s.lower())
        region_names = ["(NEW)"] + region_names
        self.region_list.set_list(region_names)
        if selected is not None:
            self.region_list.select(selected)
        self.update_right_panel()
        self.region_list.select_action(self.update)

    def update_right_panel(self):
        region_name = self.region_list.selected()
        if region_name is None or region_name == "(NEW)":
            self.name_editor.text("")
        else:
            self.name_editor.text(region_name)
        region_info = regions.get(region_name, default_region)
        self.nrnregion_selector.select(region_info.get("nrn_region", None))
        self.geoselector.select_by_name(region_info.get("geometry", "Inside"))

    def save(self):
        region_name = self.name_editor.text()
        if not region_name.strip():
            h.continue_dialog("A name is required.")
            return
        if region_name in regions:
            morphology = regions[region_name].get("morphology", "No Sections")
        else:
            morphology = "No Sections"
        regions[region_name] = {
            "morphology": morphology,
            "geometry": self.geoselector.text[self.geoselector.option],
        }
        geo_type_id = self.geoselector.option - 1
        regions[region_name].update(self.option_panels[geo_type_id].get_options())
        regions[region_name]["nrn_region"] = self.nrnregion_selector.selected()
        _the_rxd_builder.update()
        self.region_list.select(region_name)

    def delete(self):
        region_name = self.region_list.selected()
        if region_name in regions:
            del regions[region_name]
            _the_rxd_builder.update()


class InstantiatePane:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        h.xpanel("")
        h.xlabel("You may want to save your work (File -> Save Session) before")
        h.xlabel("instantiating,  as there is currently no way to make changes")
        h.xlabel("after the RxD model has been instantiated.")
        h.xpanel()
        self.spacer = h.TextEditor("", 1, 30)
        self.spacer.readonly(1)
        self.spacer.map()
        h.xpanel("", 1)
        h.xlabel("When you are ready, click:   ")
        h.xbutton("Instantiate", _instantiate)
        h.xpanel()
        self.vbox.intercept(0)
        self.vbox.map()


class _RxDBuilder:
    def __init__(self):
        global rxd_builder_tab

        self.vbox = h.VBox(3)
        self.vbox.intercept(1)

        if rxd_builder_tab not in [1, 2, 3, 4, 5]:
            rxd_builder_tab = 1

        h.xpanel("", 1)
        h.xradiobutton("Regions    ", self.regions, 1 if rxd_builder_tab == 1 else 0)
        h.xradiobutton("Species    ", self.species, 1 if rxd_builder_tab == 2 else 0)
        h.xradiobutton(
            "Reactions    ", self.reactions, 1 if rxd_builder_tab == 3 else 0
        )
        h.xradiobutton(
            "Morphology    ", self.morphology, 1 if rxd_builder_tab == 4 else 0
        )
        h.xradiobutton(
            "Instantiate", self.instantiate, 1 if rxd_builder_tab == 5 else 0
        )
        h.xpanel()

        self.deck = h.Deck()
        self.deck.intercept(1)
        self.regionpane = RegionPane()
        self.speciespane = SpeciesPane()
        self.reactionpane = ReactionPane()
        self.morphologypane = MorphologyPane()
        self.instantiatepane = InstantiatePane()
        self._the_reaction_editor = _ReactionEditor()
        self.deck.intercept(0)
        self.deck.map()

        self.deck.flip_to(rxd_builder_tab - 1)

        self.vbox.save(self.save)

        self.vbox.intercept(0)
        self.vbox.full_request(1)

    @property
    def is_mapped(self):
        return self.vbox.ismapped()

    def map(self, name=None, left=None, top=None, width=None, height=None):
        if self.is_mapped:
            return
        if name is None:
            if rxd_builder_left is None:
                self.vbox.map("RxD Builder")
            else:
                self.vbox.map(
                    "RxD Builder",
                    rxd_builder_left,
                    rxd_builder_top,
                    rxd_builder_width,
                    rxd_builder_height,
                )
        else:
            self.vbox.map(name, left, top, width, height)

    def save(self):
        self.vbox.save('nrnpython("import neuron.rxd.gui")')
        self.vbox.save('nrnpython("neuron.rxd.gui.regions = %r")' % regions)
        self.vbox.save('nrnpython("neuron.rxd.gui.species = %r")' % species)
        self.vbox.save('nrnpython("neuron.rxd.gui.all_reactions = %r")' % all_reactions)
        self.vbox.save(
            'nrnpython("neuron.rxd.gui.rxd_builder_tab = %r")' % rxd_builder_tab
        )
        self.vbox.save(
            'nrnpython("neuron.rxd.gui.has_instantiated = %r")' % has_instantiated
        )
        self.vbox.save('nrnpython("from neuron import h")')
        self.vbox.save(
            'nrnpython("h.ocbox_ = neuron.rxd.gui.RxDBuilder(visible=False)")'
        )

        if has_instantiated:
            self.vbox.save('nrnpython("h.ocbox_.instantiate()")')

    def regions(self):
        global rxd_builder_tab
        rxd_builder_tab = 1
        self.deck.flip_to(0)

    def species(self):
        global rxd_builder_tab
        rxd_builder_tab = 2
        self.deck.flip_to(1)
        self.vbox.full_request(1)

    def reactions(self):
        global rxd_builder_tab
        rxd_builder_tab = 3
        self.deck.flip_to(2)

    def morphology(self):
        global rxd_builder_tab
        rxd_builder_tab = 4
        self.deck.flip_to(3)
        self.morphologypane.update()

    def instantiate(self):
        global rxd_builder_tab
        rxd_builder_tab = 5
        self.deck.flip_to(4)
        # self.instantiatepane.update()

    def update(self):
        self.morphologypane.update()
        self.regionpane.update()
        self.speciespane.update()
        self._the_reaction_editor.update()


_the_rxd_builder = None


class NrnRegionSelector:
    def __init__(self):
        h.xpanel("NrnRegionSelector", 1)
        h.xlabel("Electrophysiology region: ")
        self.opt1, self.opt2, self.opt3 = 0, 0, 1
        h.xcheckbox("(I)nside  ", (self, "opt1"), self._click_inside)
        h.xcheckbox("(O)utside  ", (self, "opt2"), self._click_outside)
        h.xcheckbox("Neither", (self, "opt3"), self._click_neither)
        h.xpanel()

    def _click_inside(self):
        self.opt1, self.opt2, self.opt3 = 1, 0, 0
        self.region = "i"

    def _click_outside(self):
        self.opt1, self.opt2, self.opt3 = 0, 1, 0
        self.region = "o"

    def _click_neither(self):
        self.opt1, self.opt2, self.opt3 = 0, 0, 1
        self.region = None

    def selected(self):
        return self.region

    def select(self, name):
        if name == "i":
            self._click_inside()
        elif name == "o":
            self._click_outside()
        else:
            self._click_neither()


class SectionListSelector(_PartialSelector):
    def __init__(self):
        self.ell = h.List()
        self.ell.browser("title", "s")
        self.update()
        self.ell.select(-1)

    def update(self, select=None):
        was_selected = self.selected()
        self.ell.remove_all()
        self.names, self.mapping = get_sectionlists()
        self.names = ["No Sections", "All Sections"] + self.names
        for name in self.names:
            self._append(name)
        # restore selection if possible
        if was_selected is None:
            self.ell.select(-1)
        else:
            self.select(was_selected)

    def _append(self, name):
        self.ell.append(h.String(name))


class SpeciesSelectorWithRegions(_PartialSelector):
    def __init__(self, allow_without_region, allow_with_region):
        self.allow_without_region = allow_without_region
        self.allow_with_region = allow_with_region
        self.ell = h.List()
        self.ell.browser("title", "s")
        self.update()

    def update(self):
        self.ell.remove_all()
        self.names = []
        for s in species:
            if self.allow_without_region:
                self.names.append(s)
            if self.allow_with_region:
                for r in species[s]["regions"]:
                    self.names.append("%s[%s]" % (s, r))
        self.names = sorted(self.names, key=lambda s: s.lower())
        for name in self.names:
            self._append(name)

    def _append(self, name):
        self.ell.append(h.String(name))


class MembraneSelector(_PartialSelector):
    def __init__(self):
        self.ell = h.List()
        self.ell.browser("title", "s")
        self.update()

    def update(self):
        selected = self.selected()
        boundary_types = ("Membrane", "Constant 2D Area/Length")
        self.ell.remove_all()
        self.names = [
            name
            for name, data in zip(list(regions.keys()), list(regions.values()))
            if data["geometry"] in boundary_types
        ]
        self.names = sorted(self.names, key=lambda s: s.lower())
        for name in self.names:
            self._append(name)
        self.select(selected)

    def _append(self, name):
        self.ell.append(h.String(name))


class RegionList(_PartialSelector):
    def __init__(self, title, names=[]):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        if title is not None:
            h.xpanel("regionlist")
            h.xlabel(title)
            h.xpanel()
        self.ell = h.List()
        self.set_list(names)
        self.ell.browser("title", "s")
        self.vbox.intercept(0)
        self.vbox.map()

    def append(self, name):
        self.ell.append(h.String(name))
        self.names.append(name)

    def remove(self, name):
        try:
            i = self.names.index(name)
        except ValueError:
            return
        self.ell.remove(i)
        del self.names[i]

    def set_list(self, names):
        self.names = []
        self.ell.remove_all()
        for name in names:
            self.append(name)


class _MorphologyPane:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)

        h.xpanel("")
        h.xlabel("Select the morphology corresponding to each region")
        h.xpanel()

        self.hbox = h.HBox(3)
        self.hbox.intercept(1)

        self.region_list = RegionList(None)

        self.sectionlist_selector = SectionListSelector()
        self.sectionlist_selector.select_action(self.change_morphology_association)

        self.shape_plot = h.Shape()
        self.shape_plot.show(1)

        self.hbox.intercept(0)
        self.hbox.map()

        self.vbox.intercept(0)

    def map(self):
        self.vbox.map("Morphology Pane")

    def change_morphology_association(self):
        region_name = self.region_list.selected()
        morph_name = self.sectionlist_selector.selected()
        if region_name in regions:
            regions[region_name]["morphology"] = morph_name
        if morph_name == "No Sections":
            self.shape_plot.color_all(1)
        elif morph_name == "All Sections":
            self.shape_plot.color_all(2)
        else:
            self.shape_plot.color_all(1)
            self.shape_plot.color_list(
                h.SectionList[self.sectionlist_selector.mapping[morph_name]], 2
            )

    def update(self):
        # update the names
        self.region_list.select_action("")
        selected = self.region_list.selected()
        region_names = sorted(list(regions.keys()), key=lambda s: s.lower())
        self.region_list.set_list(region_names)
        if selected is not None:
            self.region_list.select(selected)
        self.update_right_panel()
        self.region_list.select_action(self.update)
        self.sectionlist_selector.select_action("")
        self.sectionlist_selector.update()
        self.sectionlist_selector.select_action(self.change_morphology_association)
        self.change_morphology_association()

    def update_right_panel(self):
        if self.region_list.selected() is None:
            if self.region_list.names:
                self.region_list.select(self.region_list.names[0])
        if self.region_list.selected() is not None:
            region = regions[self.region_list.selected()]
            self.sectionlist_selector.select(region.get("morphology", "No Sections"))
        else:
            self.sectionlist_selector.select("No Sections")


class RegionSelector:
    def __init__(self, name):
        self.name = name
        selected = species[name]["regions"]
        self._setup_left_right(selected)
        self.region_selector = h.HBox(3)
        self.region_selector.intercept(1)
        self.left_list = RegionList("Nonselected Regions")
        self.arrow_col = h.VBox(3)
        self.arrow_col.intercept(1)
        self.arrow_panel = h.xpanel("")
        h.xlabel("")
        h.xlabel("")
        self.right_arrow = h.xbutton(">", self._do_right_arrow)
        self.left_arrow = h.xbutton("<", self._do_left_arrow)
        h.xlabel("")
        h.xlabel("")
        h.xpanel()
        self.arrow_col.intercept(0)
        self.arrow_col.map()
        self.right_list = RegionList("Selected Regions")
        self.region_selector.intercept(0)
        self.region_selector.map()
        self._update_lists()

    def _update_lists(self):
        self.left_list.set_list(self.left_list_items)
        self.right_list.set_list(self.right_list_items)

    def _update_regions(self):
        species[self.name]["regions"] = self.right_list.names

    def _setup_left_right(self, right):
        left = list(regions.keys())
        right = list(right)

        # only keep those things that belong to the full region names
        # but don't put anything kept on the left
        for item in right:
            if item not in left:
                right.remove(item)
            else:
                left.remove(item)

        self.left_list_items = left
        self.right_list_items = right

    def _do_left_arrow(self):
        name = self.right_list.selected()
        if name is None:
            return
        self.right_list.remove(name)
        self.left_list.append(name)
        self._update_regions()

    def _do_right_arrow(self):
        name = self.left_list.selected()
        if name is None:
            return
        self.left_list.remove(name)
        self.right_list.append(name)
        self._update_regions()

    def selected(self):
        return list(self.right_list.names)


class SpeciesLocator:
    def __init__(self, name):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        h.xpanel("")
        h.xlabel(name)
        h.xpanel()
        self.regions_selector = RegionSelector(name)
        self.vbox.intercept(0)
        self.vbox.map()
        self.name = name

    def regions(self):
        return self.regions_selector.selected()


class SpeciesPanel:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.species_locs = {}
        for i, name in enumerate(species.keys()):
            if i:
                h.xpanel("")
                h.xlabel("")
                h.xpanel()
            self.species_locs[name] = SpeciesLocator(name)
        self.vbox.intercept(0)
        self.vbox.map()

    def regions(self):
        return dict(
            [
                (name, loc.regions())
                for name, loc in zip(
                    list(self.species_locs.keys()), list(self.species_locs.values())
                )
            ]
        )


class _SpeciesEditor:
    def __init__(self):
        self.hbox = h.HBox(3)
        self.hbox.save("")
        self.hbox.intercept(1)
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.ell = h.List()
        self._set_list()
        self.ell.browser("", "s")
        self.ell.select(0)
        h.xpanel("")
        h.xbutton("Delete", self._delete)
        h.xpanel()
        self.vbox.intercept(0)
        self.vbox.map()
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)

        h.xpanel("")
        h.xlabel("Name")
        h.xpanel()

        self.vbox.adjuster(15)
        self.name_editor = h.TextEditor("", 1, 30)
        self.name_editor.map()

        self._set_values("", 0, 0)

        h.xpanel("")
        h.xvalue("Charge", (self, "charge"))
        h.xvalue("Diff Const", (self, "d"))
        h.xpanel()

        h.xpanel("", 1)
        h.xbutton("Revert", self._revert)
        h.xbutton("Save", self._save)
        h.xpanel()

        self.vbox.intercept(0)
        self.vbox.map()
        self.hbox.intercept(0)

        self.ell.select_action(self._new_or_display, 1)

    def _new_or_display(self):
        if self.selected() is None:
            self._set_values("", 0, 0)
        else:
            self._revert()

    def _set_values(self, name, charge, diff):
        self.charge = charge
        self.d = diff
        self.name_editor.text(name)

    def _revert(self):
        name = self.selected()
        if name is not None:
            data = species[name]
            self._set_values(name, data["charge"], data["d"])
        else:
            h.continue_dialog("Nothing to revert to")

    def _save(self):
        name = self.name_editor.text()
        if not name:
            h.continue_dialog("Must enter a name to save")
        else:
            if name in species:
                regions = species[name]["regions"]
            else:
                regions = []
            species[name] = {"charge": self.charge, "d": self.d, "regions": regions}
            self._set_list()
            self._select(name)

    def _delete(self):
        name = self.selected()
        if name is not None:
            del species[name]
            self._set_list()
        else:
            h.continue_dialog("Nothing to delete")
        self._new_or_display()

    def _set_list(self, do_update=True):
        self.ell.remove_all()
        self.append("(NEW)")
        self._names = list(species.keys())
        self._names.sort()
        for name in self._names:
            self.append(name)
        if do_update:
            try:
                species_pane_update()
            except NameError:
                pass

    def append(self, name):
        self.ell.append(h.String(name))

    def _select(self, name):
        self.ell.select(1 + self._names.index(name))

    def selected(self):
        item = int(self.ell.selected())
        if item == -1 or item == 0:
            return None
        return self._names[item - 1]

    @property
    def is_mapped(self):
        return self.hbox.ismapped()

    def map(self):
        self.hbox.map("Species Editor")


_the_species_editor = _SpeciesEditor()

# this is a way of faking the Singleton pattern
def SpeciesEditor():
    if not _the_species_editor.is_mapped:
        _the_species_editor.map()
    _the_species_editor._set_list(do_update=False)
    # select (NEW)
    _the_species_editor.ell.select(0)


class _SpeciesPane:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        h.xpanel("")
        h.xlabel("Edit Species and Select Regions for Each Species")
        h.xbutton("Species Editor", SpeciesEditor)
        h.xpanel()
        self.deck = h.Deck()
        self.deck.intercept(1)
        self.speciespanel = SpeciesPanel()
        self.deck.intercept(0)
        self.deck.map()
        self.deck.flip_to(0)
        self.vbox.intercept(0)

    def _update_panel(self):
        self.deck.remove_last()
        self.deck.intercept(1)
        self.speciespanel = SpeciesPanel()
        self.deck.intercept(0)
        self.deck.flip_to(0)

    def update(self):
        self._update_panel()

    @property
    def is_mapped(self):
        return self.vbox.ismapped()

    def map(self):
        self.vbox.map("Species Pane")


_the_species_pane = _SpeciesPane()

# this is a way of faking the Singleton pattern
def SpeciesPane():
    _the_species_pane = _SpeciesPane()
    _the_species_pane.map()
    return _the_species_pane


def species_pane_update():
    global _the_rxd_builder
    #  print 'species_pane_update'
    if _the_rxd_builder:
        deck = _the_rxd_builder.deck
        #    print deck.hname()
        deck.remove(1)
        deck.intercept(1)
        _the_rxd_builder.speciespane = SpeciesPane()
        deck.intercept(0)
        deck.move_last(1)
        deck.flip_to(rxd_builder_tab - 1)


class SpeciesMultiSelector:
    def __init__(self, selector, allow_with_region=False, allow_without_region=True):
        self.selector = selector
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.ell = h.List()
        self.names = []
        if allow_without_region:
            self.names += list(species.keys())
        if allow_with_region:
            for s in list(species.keys()):
                for r in species[s]["regions"]:
                    self.names.append("%s[%s]" % (s, r))

        self.names.sort()
        for name in self.names:
            self.ell.append(h.String(name))
        self.ell.browser("title", "s")
        self.data = numpy.zeros(1)
        h.xpanel("")
        h.xpvalue("Multiplicity", neuron.numpy_element_ref(self.data, 0))
        h.xpanel()
        h.xpanel("", 1)
        h.xbutton("Cancel", self._cancel)
        h.xbutton("Accept", self._accept)
        h.xpanel()
        self.vbox.intercept(0)
        self.vbox.map("Species MultiSelector")

    def _accept(self):
        i = int(self.ell.selected())
        mult = self.data[0]
        if i == -1 or mult <= 0:
            h.continue_dialog("Must select a Species with positive multiplicity")
            return
        self.selector.add(self.names[i], mult)
        self.vbox.unmap()

    def _cancel(self):
        self.vbox.unmap()


class LRHSSelector(_PartialSelector):
    def __init__(self, allow_without_region=True, allow_with_region=False):
        self.allow_without_region = allow_without_region
        self.allow_with_region = allow_with_region
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.names = []
        self.ell = h.List()
        self._update_list()
        self.species = []
        self.mults = []
        self.ell.browser("", "s")
        h.xpanel("", 1)
        h.xbutton("Add", self._add)
        h.xbutton("Remove", self._remove)
        h.xpanel()
        self.vbox.intercept(0)
        self.vbox.map("L/RHS Selector")

    def add(self, name, mult):
        self.species.append(name)
        if mult == int(mult):
            mult = int(mult)
        self.mults.append(mult)
        self.names.append("%g * %s" % (mult, name))
        self._update_list()

    def _add(self):
        SpeciesMultiSelector(
            self,
            allow_with_region=self.allow_with_region,
            allow_without_region=self.allow_without_region,
        )

    def _update_list(self):
        self.ell.remove_all()
        for name in self.names:
            self._append(name)

    def _append(self, name):
        self.ell.append(h.String(name))

    def _remove(self):
        self.remove(self.selected())

    def remove(self, name):
        try:
            i = self.names.index(name)
        except ValueError:
            return
        self.ell.remove(i)
        del self.names[i]
        del self.species[i]
        del self.mults[i]
        self._update_list()

    def update(self, data):
        self.species = []
        self.mults = []
        self.names = []
        if data:
            for name, mult in data:
                self.add(name, mult)
        else:
            self.ell.remove_all()

    def selection(self):
        return [(name, mult) for name, mult in zip(self.names, self.mults)]


class SpecificRateEditor:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.selector = SpeciesSelectorWithRegions(True, True)
        h.xpanel("")
        h.xlabel("Production Rate [units: mM/ms]:")
        h.xpanel()
        self.vbox.adjuster(15)
        self.kf_editor = h.TextEditor("0", 1, 30)
        self.kf_editor.map()
        self.vbox.intercept(0)
        self.vbox.map()

    def update(self, data):
        self.kf_editor.text(data["kf"])
        self.selector.update()
        self.selector.select(data.get("species", ""))

    def get_options(self):
        return {
            "type": "rate",
            "kf": self.kf_editor.text(),
            "species": self.selector.selected(),
        }


def _do_nothing():
    pass


class SpecificReactionEditor:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.hbox = h.HBox(3)
        self.hbox.intercept(1)
        self.lhs_selector = LRHSSelector()
        h.xpanel("")
        h.xlabel("")
        h.xlabel("")
        h.xbutton("kf -->", _do_nothing)
        h.xbutton("<-- kb", _do_nothing)
        h.xpanel()
        self.rhs_selector = LRHSSelector()
        self.hbox.intercept(0)
        self.hbox.map("Specific Reaction Editor")

        h.xpanel("")
        h.xlabel("kf:")
        h.xpanel()
        self.vbox.adjuster(15)

        self.kf_editor = h.TextEditor("0", 1, 30)
        self.kf_editor.map()

        h.xpanel("")
        h.xlabel("kb:")
        h.xpanel()
        self.vbox.adjuster(15)

        self.kb_editor = h.TextEditor("0", 1, 30)
        self.kb_editor.map()

        self.is_massaction = True
        h.xpanel("")
        h.xcheckbox("Mass Action", (self, "is_massaction"))
        h.xpanel()

        self.vbox.intercept(0)
        self.vbox.map()

    def update(self, data):
        self.is_massaction = data.get("massaction", False)
        self.kf_editor.text(data["kf"])
        self.kb_editor.text(str(data.get("kb", "0")))
        self.lhs_selector.update(data.get("lhs", []))
        self.rhs_selector.update(data.get("rhs", []))

    def get_options(self):
        return {
            "type": "reaction",
            "kf": self.kf_editor.text(),
            "kb": self.kb_editor.text(),
            "lhs": self.lhs_selector.selection(),
            "rhs": self.rhs_selector.selection(),
        }


class SpecificMultiCompartmentReactionEditor:
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.hbox = h.HBox(3)
        self.hbox.intercept(1)
        self.lhs_selector = LRHSSelector(
            allow_with_region=True, allow_without_region=False
        )
        h.xpanel("")
        h.xlabel("")
        h.xlabel("")
        h.xbutton("kf -->", _do_nothing)
        h.xbutton("<-- kb", _do_nothing)
        h.xpanel()
        self.rhs_selector = LRHSSelector()
        self.hbox.intercept(0)
        self.hbox.map("Specific Reaction Editor")

        h.xpanel("")
        h.xlabel("kf:")
        h.xpanel()

        self.vbox.adjuster(15)
        self.kf_editor = h.TextEditor("0", 1, 30)
        self.kf_editor.map()

        h.xpanel("")
        h.xlabel("kb:")
        h.xpanel()

        self.vbox.adjuster(15)
        self.kb_editor = h.TextEditor("0", 1, 30)
        self.kb_editor.map()

        h.xpanel("")
        h.xlabel("Select the Boundary:")
        h.xpanel()
        self.membrane_selector = MembraneSelector()

        self.is_massaction = True
        self.membrane_current = False
        self.scale_with_area = True
        h.xpanel("")
        h.xcheckbox("Mass Action", (self, "is_massaction"))
        h.xcheckbox("Induces Membrane Current", (self, "membrane_current"))
        h.xcheckbox("Scales With Membrane Area", (self, "scale_with_area"))
        h.xpanel()

        self.vbox.intercept(0)
        self.vbox.map()

    def get_options(self):
        return {
            "type": "multicompartmentreaction",
            "kf": self.kf_editor.text(),
            "kb": self.kb_editor.text(),
            "lhs": self.lhs_selector.selection(),
            "rhs": self.rhs_selector.selection(),
            "membrane": self.membrane_selector.selected(),
        }

    def update(self, data):
        self.membrane_selector.update()
        self.is_massaction = data.get("massaction", False)
        self.kf_editor.text(data["kf"])
        self.kb_editor.text(str(data.get("kb", "0")))
        self.lhs_selector.update(data.get("lhs", []))
        self.rhs_selector.update(data.get("rhs", []))
        self.scale_with_area = data.get("scale_with_area", True)
        self.membrane_current = data.get("membrane_current", False)
        self.membrane_selector.select(data.get("membrane", None))


class ReactionPanelRight:
    def __init__(self):
        self.data = {
            "type": "reaction",
            "kf": "0",
            "kb": "0",
            "massaction": True,
            "lhs": [],
            "rhs": [],
        }
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        h.xpanel("")
        h.xlabel("Name:")
        h.xpanel()
        self.vbox.adjuster(15)
        self.name_editor = h.TextEditor("", 1, 30)
        self.name_editor.map()
        self.opt1 = 0
        self.opt2 = 1
        self.opt3 = 0
        h.xpanel("", 1)
        h.xcheckbox("ODE   ", (self, "opt1"), self._select_rate)
        h.xcheckbox("Reaction   ", (self, "opt2"), self._select_reaction)
        h.xcheckbox(
            "MultiCompartmentReaction   ",
            (self, "opt3"),
            self._select_multicompartmentreaction,
        )
        h.xpanel()
        self.deck = h.Deck()
        self.deck.intercept(1)
        self.rate_editor = SpecificRateEditor()
        self.reaction_editor = SpecificReactionEditor()
        self.multicompartmentreaction_editor = SpecificMultiCompartmentReactionEditor()
        self.deck.intercept(0)
        self.deck.map()
        self.deck.flip_to(1)
        self.current_editor = self.reaction_editor
        h.xpanel("", 1)
        h.xbutton("Revert", self._revert)
        h.xbutton("Save", self._save)
        h.xpanel()
        self.vbox.intercept(0)
        self.vbox.map()

    def _save(self):
        name = self.name_editor.text()
        if not name:
            h.continue_dialog("Must enter a name to save")
        else:
            reaction = self.current_editor.get_options()
            if name in all_reactions:
                reaction["active"] = all_reactions[name]["active"]
            else:
                reaction["active"] = False
            all_reactions[name] = reaction
        _the_rxd_builder.update()
        _the_rxd_builder._the_reaction_editor._select(name)
        _the_rxd_builder._the_reaction_editor._update_view()

    def _revert(self):
        _the_rxd_builder._the_reaction_editor._update_view()

    def _select_rate(self):
        self.opt1, self.opt2, self.opt3 = 1, 0, 0
        self.deck.flip_to(0)
        self.rate_editor.update(self.data)
        self.current_editor = self.rate_editor

    def _select_reaction(self):
        self.deck.flip_to(1)
        self.opt1, self.opt2, self.opt3 = 0, 1, 0
        self.reaction_editor.update(self.data)
        self.current_editor = self.reaction_editor

    def _select_multicompartmentreaction(self):
        self.deck.flip_to(2)
        self.opt1, self.opt2, self.opt3 = 0, 0, 1
        self.multicompartmentreaction_editor.update(self.data)
        self.current_editor = self.multicompartmentreaction_editor

    def update(self, name, data):
        if name is None:
            name = ""
        self.name_editor.text(name)
        self.data = data
        kind = data["type"]
        if kind == "rate":
            self._select_rate()
        elif kind == "reaction":
            self._select_reaction()
        elif kind == "multicompartmentreaction":
            self._select_multicompartmentreaction()

    def update2(self):
        self.multicompartmentreaction.update()


class _ReactionPane(object):
    def __init__(self):
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        h.xpanel("")
        h.xlabel("Edit and Enable Reactions")
        h.xbutton("Reaction Editor", ReactionEditor)
        h.xpanel()
        self.deck = h.Deck()
        self.deck.intercept(1)
        h.xpanel("")
        h.xpanel()
        self.deck.intercept(0)
        self.deck.map()
        self.deck.flip_to(0)
        self.vbox.intercept(0)
        self._update_panel()

    def _update_panel(self):
        self.deck.remove_last()
        self.deck.intercept(1)
        self.reaction_names = sorted(
            list(all_reactions.keys()), key=lambda s: s.lower()
        )
        for i, name in enumerate(self.reaction_names):
            self.__setattr__(
                "include_list%d" % i, 1 if all_reactions[name]["active"] else 0
            )
        h.xpanel("")
        if self.reaction_names:
            for i, name in enumerate(self.reaction_names):
                h.xcheckbox(
                    name, (self, "include_list%d" % i), self._update_active_reactions
                )
        else:
            h.xlabel("*** No reactions defined yet ***")
        h.xpanel()
        self.deck.intercept(0)
        self.deck.flip_to(0)

    def _update_active_reactions(self):
        for i, name in enumerate(self.reaction_names):
            all_reactions[name]["active"] = (
                True if getattr(self, "include_list%d" % i) else False
            )

    @property
    def is_mapped(self):
        return self.vbox.ismapped()

    def map(self):
        self.vbox.map("Reaction Pane")


_the_reaction_pane = _ReactionPane()

# this is a way of faking the Singleton pattern
def ReactionPane():
    if not _the_reaction_pane.is_mapped:
        _the_reaction_pane.map()


class _ReactionEditor:
    def __init__(self):
        self.hbox = h.HBox(3)
        self.hbox.save("")
        self.hbox.intercept(1)
        self.vbox = h.VBox(3)
        self.vbox.intercept(1)
        self.ell = h.List()
        self._set_list()
        self.ell.browser("", "s")
        self.ell.select(0)
        h.xpanel("")
        h.xbutton("Delete", self._delete)
        h.xpanel()
        self.vbox.intercept(0)
        self.vbox.map()
        self.right_panel = ReactionPanelRight()
        self.hbox.intercept(0)
        self.ell.select_action(self._update_view)
        self.hbox.full_request(1)

    def update(self):
        self._set_list()
        self._update_view()

    def _update_view(self):
        name = self.selected()
        if name is None:
            data = {
                "type": "reaction",
                "kf": "0",
                "kb": "0",
                "massaction": True,
                "lhs": [],
                "rhs": [],
            }
        else:
            data = all_reactions[name]
        self.right_panel.update(name, data)

    def _select(self, name):
        self.ell.select(1 + self._names.index(name))

    def selected(self):
        item = int(self.ell.selected())
        if item == -1 or item == 0:
            return None
        return self._names[item - 1]

    def _delete(self):
        try:
            del all_reactions[self.selected()]
        except:
            pass
        self._set_list()

    @property
    def is_mapped(self):
        return self.hbox.ismapped()

    def map(self):
        self.hbox.map("Reaction Editor")

    def _set_list(self):
        selected = self.selected()
        self.ell.remove_all()
        self._append("(NEW)")
        self._names = sorted(list(all_reactions.keys()), key=lambda s: s.lower())
        for name in self._names:
            self._append(name)
        try:
            species_pane_update()
        except NameError:
            pass
        try:
            _the_reaction_pane._update_panel()
        except NameError:
            pass
        if selected in self._names:
            self._select(selected)
        else:
            self.ell.select(0)

    def _append(self, name):
        self.ell.append(h.String(name))



from neuron import h
import math


class Cell:
    def __init__(self):
        soma = h.Section(name="soma", cell=self)
        dend = h.Section(name="dend", cell=self)
        dend.connect(soma)

        soma.nseg = 1
        soma.L = 10
        soma.diam = 100.0 / soma.L / math.pi
        soma.insert("hh")

        dend.nseg = 25
        dend.L = 1000
        dend.diam = 2
        dend.insert("hh")
        for seg in dend:
            seg.hh.gnabar /= 2
            seg.hh.gkbar /= 2
            seg.hh.gl /= 2

        self.dend = dend
        self.soma = soma
        for sec in [soma, dend]:
            sec.Ra = 100
            sec.cm = 1



#!/usr/bin/env python

"""
Locate libpython associated with this Python executable.
"""

# https://pypi.org/project/find-libpython/

# License
#
# Copyright 2018, Takafumi Arakaki
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function, absolute_import

from logging import getLogger
import ctypes.util
import functools
import os
import sys
import sysconfig

logger = getLogger("find_libpython")

is_windows = os.name == "nt"
is_apple = sys.platform == "darwin"

SHLIB_SUFFIX = sysconfig.get_config_var("SHLIB_SUFFIX")
if SHLIB_SUFFIX is None:
    if is_windows:
        SHLIB_SUFFIX = ".dll"
    else:
        SHLIB_SUFFIX = ".so"
if is_apple:
    # sysconfig.get_config_var("SHLIB_SUFFIX") can be ".so" in macOS.
    # Let's not use the value from sysconfig.
    SHLIB_SUFFIX = ".dylib"


def linked_libpython():
    """
    Find the linked libpython using dladdr (in *nix).

    Calling this in Windows always return `None` at the moment.

    Returns
    -------
    path : str or None
        A path to linked libpython.  Return `None` if statically linked.
    """
    if is_windows:
        return None
    return _linked_libpython_unix()


class Dl_info(ctypes.Structure):
    _fields_ = [
        ("dli_fname", ctypes.c_char_p),
        ("dli_fbase", ctypes.c_void_p),
        ("dli_sname", ctypes.c_char_p),
        ("dli_saddr", ctypes.c_void_p),
    ]


def _linked_libpython_unix():
    libdl = ctypes.CDLL(ctypes.util.find_library("dl"))
    libdl.dladdr.argtypes = [ctypes.c_void_p, ctypes.POINTER(Dl_info)]
    libdl.dladdr.restype = ctypes.c_int

    dlinfo = Dl_info()
    retcode = libdl.dladdr(
        ctypes.cast(ctypes.pythonapi.Py_GetVersion, ctypes.c_void_p),
        ctypes.pointer(dlinfo),
    )
    if retcode == 0:  # means error
        return None
    path = os.path.realpath(dlinfo.dli_fname.decode())
    if path == os.path.realpath(sys.executable):
        return None
    return path


def library_name(name, suffix=SHLIB_SUFFIX, is_windows=is_windows):
    """
    Convert a file basename `name` to a library name (no "lib" and ".so" etc.)

    >>> library_name("libpython3.7m.so")                   # doctest: +SKIP
    'python3.7m'
    >>> library_name("libpython3.7m.so", suffix=".so", is_windows=False)
    'python3.7m'
    >>> library_name("libpython3.7m.dylib", suffix=".dylib", is_windows=False)
    'python3.7m'
    >>> library_name("python37.dll", suffix=".dll", is_windows=True)
    'python37'
    """
    if not is_windows and name.startswith("lib"):
        name = name[len("lib") :]
    if suffix and name.endswith(suffix):
        name = name[: -len(suffix)]
    return name


def append_truthy(list, item):
    if item:
        list.append(item)


def uniquifying(items):
    """
    Yield items while excluding the duplicates and preserving the order.

    >>> list(uniquifying([1, 2, 1, 2, 3]))
    [1, 2, 3]
    """
    seen = set()
    for x in items:
        if x not in seen:
            yield x
        seen.add(x)


def uniquified(func):
    """Wrap iterator returned from `func` by `uniquifying`."""

    @functools.wraps(func)
    def wrapper(*args, **kwds):
        return uniquifying(func(*args, **kwds))

    return wrapper


@uniquified
def candidate_names(suffix=SHLIB_SUFFIX):
    """
    Iterate over candidate file names of libpython.

    Yields
    ------
    name : str
        Candidate name libpython.
    """
    LDLIBRARY = sysconfig.get_config_var("LDLIBRARY")
    if LDLIBRARY:
        yield LDLIBRARY

    LIBRARY = sysconfig.get_config_var("LIBRARY")
    if LIBRARY:
        yield os.path.splitext(LIBRARY)[0] + suffix

    dlprefix = "" if is_windows else "lib"
    sysdata = dict(
        v=sys.version_info,
        # VERSION is X.Y in Linux/macOS and XY in Windows:
        VERSION=(
            sysconfig.get_config_var("VERSION")
            or "{v.major}.{v.minor}".format(v=sys.version_info)
        ),
        ABIFLAGS=(
            sysconfig.get_config_var("ABIFLAGS")
            or sysconfig.get_config_var("abiflags")
            or ""
        ),
    )

    for stem in [
        "python{VERSION}{ABIFLAGS}".format(**sysdata),
        "python{VERSION}".format(**sysdata),
        "python{v.major}".format(**sysdata),
        "python",
    ]:
        yield dlprefix + stem + suffix


@uniquified
def candidate_paths(suffix=SHLIB_SUFFIX):
    """
    Iterate over candidate paths of libpython.

    Yields
    ------
    path : str or None
        Candidate path to libpython.  The path may not be a fullpath
        and may not exist.
    """

    yield linked_libpython()

    # List candidates for directories in which libpython may exist
    lib_dirs = []
    append_truthy(lib_dirs, sysconfig.get_config_var("LIBPL"))
    append_truthy(lib_dirs, sysconfig.get_config_var("srcdir"))
    append_truthy(lib_dirs, sysconfig.get_config_var("LIBDIR"))

    # LIBPL seems to be the right config_var to use.  It is the one
    # used in python-config when shared library is not enabled:
    # https://github.com/python/cpython/blob/v3.7.0/Misc/python-config.in#L55-L57
    #
    # But we try other places just in case.

    if is_windows:
        lib_dirs.append(os.path.join(os.path.dirname(sys.executable)))
    else:
        lib_dirs.append(
            os.path.join(os.path.dirname(os.path.dirname(sys.executable)), "lib")
        )

    # For macOS:
    append_truthy(lib_dirs, sysconfig.get_config_var("PYTHONFRAMEWORKPREFIX"))

    lib_dirs.append(sys.exec_prefix)
    lib_dirs.append(os.path.join(sys.exec_prefix, "lib"))

    lib_basenames = list(candidate_names(suffix=suffix))

    for directory in lib_dirs:
        for basename in lib_basenames:
            yield os.path.join(directory, basename)

    # In macOS and Windows, ctypes.util.find_library returns a full path:
    for basename in lib_basenames:
        yield ctypes.util.find_library(library_name(basename))


# Possibly useful links:
# * https://packages.ubuntu.com/bionic/amd64/libpython3.6/filelist
# * https://github.com/Valloric/ycmd/issues/518
# * https://github.com/Valloric/ycmd/pull/519


def normalize_path(path, suffix=SHLIB_SUFFIX, is_apple=is_apple):
    """
    Normalize shared library `path` to a real path.

    If `path` is not a full path, `None` is returned.  If `path` does
    not exists, append `SHLIB_SUFFIX` and check if it exists.
    Finally, the path is canonicalized by following the symlinks.

    Parameters
    ----------
    path : str ot None
        A candidate path to a shared library.
    """
    if not path:
        return None
    if not os.path.isabs(path):
        return None
    if os.path.exists(path):
        return os.path.realpath(path)
    if os.path.exists(path + suffix):
        return os.path.realpath(path + suffix)
    if is_apple:
        return normalize_path(_remove_suffix_apple(path), suffix=".so", is_apple=False)
    return None


def _remove_suffix_apple(path):
    """
    Strip off .so or .dylib.

    >>> _remove_suffix_apple("libpython.so")
    'libpython'
    >>> _remove_suffix_apple("libpython.dylib")
    'libpython'
    >>> _remove_suffix_apple("libpython3.7")
    'libpython3.7'
    """
    if path.endswith(".dylib"):
        return path[: -len(".dylib")]
    if path.endswith(".so"):
        return path[: -len(".so")]
    return path


@uniquified
def finding_libpython():
    """
    Iterate over existing libpython paths.

    The first item is likely to be the best one.

    Yields
    ------
    path : str
        Existing path to a libpython.
    """
    logger.debug("is_windows = %s", is_windows)
    logger.debug("is_apple = %s", is_apple)
    for path in candidate_paths():
        logger.debug("Candidate: %s", path)
        normalized = normalize_path(path)
        if normalized:
            logger.debug("Found: %s", normalized)
            yield normalized
        else:
            logger.debug("Not found.")


def find_libpython():
    """
    Return a path (`str`) to libpython or `None` if not found.

    Parameters
    ----------
    path : str or None
        Existing path to the (supposedly) correct libpython.
    """
    for path in finding_libpython():
        return os.path.realpath(path)


def print_all(items):
    for x in items:
        print(x)


def cli_find_libpython(cli_op, verbose):
    import logging

    # Importing `logging` module here so that using `logging.debug`
    # instead of `logger.debug` outside of this function becomes an
    # error.

    if verbose:
        logging.basicConfig(format="%(levelname)s %(message)s", level=logging.DEBUG)

    if cli_op == "list-all":
        print_all(finding_libpython())
    elif cli_op == "candidate-names":
        print_all(candidate_names())
    elif cli_op == "candidate-paths":
        print_all(p for p in candidate_paths() if p and os.path.isabs(p))
    else:
        path = find_libpython()
        if path is None:
            return 1
        print(path, end="")


def main(args=None):
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Print debugging information."
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--list-all",
        action="store_const",
        dest="cli_op",
        const="list-all",
        help="Print list of all paths found.",
    )
    group.add_argument(
        "--candidate-names",
        action="store_const",
        dest="cli_op",
        const="candidate-names",
        help="Print list of candidate names of libpython.",
    )
    group.add_argument(
        "--candidate-paths",
        action="store_const",
        dest="cli_op",
        const="candidate-paths",
        help="Print list of candidate paths of libpython.",
    )

    ns = parser.parse_args(args)
    parser.exit(cli_find_libpython(**vars(ns)))


if __name__ == "__main__":
    main()



# Python 3 only
# ------------------------------------------------------------------------------
# class factory for subclassing h.anyclass
# h.anyclass methods may be overridden. If so the base method can be called
# using the idiom self.basemethod = self.baseattr('methodname')
# ------------------------------------------------------------------------------

__all__ = ["hclass", "HocBaseObject"]

from . import h, hoc
import nrn
import sys


def assert_not_hoc_composite(cls):
    """
    Asserts that a class is not directly composed of multiple HOC types.
    """
    hoc_bases = set(
        b._hoc_type
        for b in cls.__bases__
        if issubclass(b, HocBaseObject) and b is not HocBaseObject
    )
    if len(hoc_bases) > 1:
        bases = ", ".join(b.__name__ for b in cls.__bases__)
        raise TypeError(f"Composition of {bases} HocObjects with different HOC types.")


def _overrides(cls, base, method_name):
    return getattr(cls, method_name) is not getattr(base, method_name)


def hclass(hoc_type, module_name=None, name=None):
    """
    Class factory for subclassing HOC types.

    Example:

    ..code-block:: python

        import neuron

        myClassTemplate = neuron.hclass(
            neuron.h.Vector,
            module_name=__name__,
            name="MyVector",
        )

        class MyVector(myClassTemplate):
            pass

    :param hoc_type: HOC types/classes such as ``h.List``, ``h.NetCon``, ``h.Vector``, ...
    :type hoc_type: :class:`hoc.HocObject`
    :param module_name: Name of the module where the class will be stored
        (usually ``__name__``)
    :type module_name: str
    :param name: Name of the module level variable that the class will be stored in. When
        omitted the name of the HOC type is used.
    :deprecated: Inherit from :class:`~neuron.HocBaseObject` instead.
    """
    if hoc_type == h.Section:
        return nrn.Section
    if module_name is None:
        module_name = __name__
    if name is None:
        name = hoc_type.hname()[:-2]
    try:
        hc = type(name, (HocBaseObject,), {}, hoc_type=hoc_type)
    except TypeError:
        raise TypeError("Argument is not a valid HOC type.") from None
    hc.__module__ = module_name
    hc.__name__ = name
    hc.__qualname__ = name
    return hc


class HocBaseObject(hoc.HocObject):
    """
    The base class for inheriting from HOC types.

    .. code-block:: python

        import neuron

        class MyVector(neuron.HocBaseObject, hoc_type=neuron.h.Vector):
            pass

    The ``__new__`` method passes the hoc type to the builtin ``hoc.HocObject`` and
    returns a new instance with the correct mapping_proxy based on the hoc type template.
    """

    def __init_subclass__(cls, hoc_type=None, **kwargs):
        if hoc_type is not None:
            if not isinstance(hoc_type, hoc.HocObject):
                raise TypeError(
                    f"Class's `hoc_type` {hoc_type} is not a valid HOC type."
                )
            cls._hoc_type = hoc_type
        elif not hasattr(cls, "_hoc_type"):
            raise TypeError(
                "Class keyword argument `hoc_type` is required for HocBaseObjects."
            )
        assert_not_hoc_composite(cls)
        hobj = hoc.HocObject
        hbase = HocBaseObject
        if _overrides(cls, hobj, "__init__") and not _overrides(cls, hbase, "__new__"):
            # Subclasses that override `__init__` must also implement `__new__` to deal
            # with the arguments that have to be passed into `HocObject.__new__`.
            # See https://github.com/neuronsimulator/nrn/issues/1129
            raise TypeError(
                f"`{cls.__qualname__}` implements `__init__` but misses `__new__`. "
                + "Class must implement `__new__`"
                + " and call `super().__new__` with the arguments required by HOC"
                + f" to construct the underlying h.{cls._hoc_type.hname()} HOC object."
            )
        super().__init_subclass__(**kwargs)

    def __new__(cls, *args, **kwds):
        # To construct HOC objects within NEURON from the Python interface, we use the
        # C-extension module `hoc`. `hoc.HocObject.__new__` both creates an internal
        # representation of the object in NEURON, and hands us back a Python object that
        # is linked to that internal representation. The `__new__` functions takes the
        # arguments that HOC objects of that type would take, and uses the `hocbase`
        # keyword argument to determine which type of HOC object to create. The `sec`
        # keyword argument can be passed along in case the construction of a HOC object
        # requires section stack access.
        kwds2 = {"hocbase": cls._hoc_type}
        if "sec" in kwds:
            kwds2["sec"] = kwds["sec"]
        return hoc.HocObject.__new__(cls, *args, **kwds2)
